#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Drops that interact with S3
"""
from asyncio.log import logger
from io import BytesIO
from typing import Tuple

from overrides import overrides

try:
    import boto3
    import botocore

except ImportError:
    logger.warning("BOTO bindings are not available")

from .data_base import DataDROP
from dlg.data.io import ErrorIO, OpenMode, DataIO
from dlg.meta import (
    dlg_component,
    dlg_batch_input,
    dlg_batch_output,
    dlg_streaming_input,
    dlg_string_param,
    dlg_list_param,
)

# from dlg.named_port_utils import identify_named_ports, check_ports_dict


##
# @brief S3
# @details An object available in a bucket on a S3 (Simple Storage Service) object storage platform
# @par EAGLE_START
# @param category S3
# @param tag daliuge
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param Bucket /String/ComponentParameter/NoPort/ReadWrite//False/False/The S3 Bucket
# @param Key /String/ComponentParameter/NoPort/ReadWrite//False/False/The S3 object key
# @param profile_name /String/ComponentParameter/NoPort/ReadWrite//False/False/The S3 profile name
# @param endpoint_url /String/ComponentParameter/NoPort/ReadWrite//False/False/The URL exposing the S3 REST API
# @param block_skip False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/If set the drop will block a skipping chain until the last producer has finished and is not also skipped.
# @param dropclass dlg.data.drops.s3_drop.S3DROP/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param base_name s3_drop/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param streaming False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component streams input and output data
# @param io /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Input Output port
# @par EAGLE_END
class S3DROP(DataDROP):
    """
    A DROP that points to data stored in S3
    """

    component_meta = dlg_component(
        "S3DROP",
        "S3 Data Drop",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    Bucket = dlg_string_param("Bucket", None)
    Key = dlg_string_param("Key", None)
    storage_class = dlg_string_param("storage_class", "S3")
    tags = dlg_list_param("tags", None)
    # don't change the aws names
    aws_access_key_id = dlg_string_param("aws_access_key_id", None)
    aws_secret_access_key = dlg_string_param("aws_secret_access_key", None)
    profile_name = dlg_string_param("profile_name", None)
    endpoint_url = dlg_string_param("endpoint_url", None)

    def initialize(self, **kwargs):
        self.keyargs = {
            "Bucket": self.Bucket,
            "Key": self.Key,
            "storage_class": self.storage_class,
            "tags": self.tags,
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "profile_name": self.profile_name,
            "endpoint_url": self.endpoint_url,
        }
        logger.debug("S3 initializing: %s", self.keyargs)
        self.Key = self.uid if not self.Key else self.Key
        return super().initialize(**kwargs)

    @property
    def path(self) -> str:
        """
        Returns the path to the S3 object
        :return: the path
        """
        return "{}/{}".format(self.Bucket, self.Key)

    @property
    def dataURL(self) -> str:
        return "s3://{}/{}".format(self.Bucket, self.Key)

    @property
    def size(self) -> int:
        size = self.getIO()._size()
        logger.debug(("Size of object: %s", size))
        return size

    def getIO(self) -> DataIO:
        """
        Return
        :return:
        """
        logger.debug("S3DROP producers: %s", self._producers)
        # if check_ports_dict(self._producers):
        #     self.mapped_inputs = identify_named_ports(
        #         self._producers, {}, self.keyargs, mode="inputs"
        #     )
        logger.debug("Parameters found: %s", self.parameters)
        return S3IO(
            self.aws_access_key_id,
            self.aws_secret_access_key,
            self.profile_name,
            self.Bucket,
            self.Key,
            self.endpoint_url,
            self._expectedSize,
        )


class S3IO(DataIO):
    """
    IO class for the S3 Drop
    """

    _desc = None

    def __init__(
        self,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        profile_name=None,
        Bucket=None,
        Key=None,
        endpoint_url=None,
        expectedSize=-1,
        **kwargs,
    ):
        super().__init__(**kwargs)

        logger.debug(
            (
                "key_id: %s; key: %s; profile: %s; bucket: %s; object_id: %s; %s",
                aws_access_key_id,
                aws_secret_access_key,
                profile_name,
                Bucket,
                Key,
                endpoint_url,
            )
        )
        self._s3 = None
        self._s3_access_key_id = aws_access_key_id
        self._s3_secret_access_key = aws_secret_access_key
        self._s3_endpoint_url = endpoint_url
        self._profile_name = profile_name
        self._bucket = Bucket
        self._key = Key
        self._s3_endpoint_url = endpoint_url
        self._s3 = self._get_s3_connection()
        self.url = f"{endpoint_url}/{Bucket}/{Key}"
        self._expectedSize = expectedSize
        self._buffer = b""
        if self._mode == 1:
            try:
                self._s3Stream = self._open()
            except botocore.exceptions.ClientError:
                if not self.exists():
                    logger.debug("Object does not exist yet. Creating!")
                    self._mode = 0

    def _get_s3_connection(self):
        s3 = None
        if self._s3 is None:
            if self._profile_name is not None or (
                self._s3_access_key_id is not None
                and self._s3_secret_access_key is not None
            ):
                logger.debug("Opening boto3 session")
                session = boto3.Session(profile_name=self._profile_name)
                s3 = session.client(
                    service_name="s3",
                    endpoint_url=self._s3_endpoint_url,
                )
            else:
                s3 = boto3.resource("s3")
        else:
            s3 = self._s3
        return s3

    def _open(self, **kwargs):
        logger.debug("Opening S3 object %s in mode %s", self._key, self._mode)
        if self._mode == OpenMode.OPEN_WRITE:
            exists = self._exists()
            if exists == (True, True):
                logger.error("Object exists already. Assuming part upload.")

            elif not exists[0]:
                # bucket does not exist, create first
                try:
                    self._s3.create_bucket(Bucket=self._bucket)
                except botocore.exceptions.ClientError as e:
                    raise e
            resp = self._s3.create_multipart_upload(
                Bucket=self._bucket,
                Key=self._key,
            )
            self._uploadId = resp["UploadId"]
            self._buffer = b""
            self._written = 0
            self._partNo = 1
            self._parts = {"Parts": []}
            return self._s3
        else:
            s3Object = self._s3.get_object(Bucket=self._bucket, Key=self._key)
            self._desc = s3Object["Body"]
        return s3Object["Body"]

    @overrides
    def _read(self, count=-1, **kwargs):
        # Read data from S3 and give it back to our reader
        if not self._desc:
            self._desc = self._open()
        if count != -1:
            return self._desc.read(count)
        else:
            return self._desc.read()

    def _writeBuffer2S3(self, write_buffer=b""):
        try:
            with BytesIO(write_buffer) as f:
                self._s3.upload_part(
                    Body=f,
                    Bucket=self._bucket,
                    Key=self._key,
                    UploadId=self._uploadId,
                    PartNumber=self._partNo,
                )
            logger.debug(
                "Wrote %d bytes part %d to S3: %s",
                len(write_buffer),
                self._partNo,
                self.url,
            )
            self._partNo += 1
            self._written += len(write_buffer)
        except botocore.exceptions.ClientError:
            logger.error("Writing to S3 failed")
            return -1

    @overrides
    def _write(self, data, **kwargs) -> int:
        """ """
        self._buffer += data
        PART_SIZE = 5 * 1024**2
        logger.debug("Length of S3 buffer: %d", len(self._buffer))
        if len(self._buffer) >= PART_SIZE:
            self._writeBuffer2S3(self._buffer[:PART_SIZE])
            self._buffer = self._buffer[PART_SIZE:]
        return len(data)  # we return the length of what we have received
        # to keep the client happy

    def _get_object_head(self) -> dict:
        return self._s3.head_object(Bucket=self._bucket, Key=self._key)

    @overrides
    def _size(self, **kwargs) -> int:
        if self.exists():
            object_head = self._get_object_head()
            logger.debug(("Size of object:%s", object_head["ContentLength"]))
            return object_head["ContentLength"]
        return -1

    @overrides
    def _close(self, **kwargs):
        if self._mode == OpenMode.OPEN_WRITE:
            if (
                len(self._buffer) > 0
            ):  # write, if there is still something in the buffer
                self._writeBuffer2S3(self._buffer)
            # complete multipart upload and cleanup
            res = self._s3.list_parts(
                Bucket=self._bucket, Key=self._key, UploadId=self._uploadId
            )
            parts = [
                {"ETag": p["ETag"], "PartNumber": p["PartNumber"]} for p in res["Parts"]
            ]
            # TODO: Check checksum!
            res = self._s3.complete_multipart_upload(
                Bucket=self._bucket,
                Key=self._key,
                UploadId=self._uploadId,
                MultipartUpload={"Parts": parts},
            )
            del self._buffer
            logger.info(
                "Wrote a total of %.1f MB to %s",
                self._written / (1024**2),
                self.url,
            )

        self._desc.close()
        del self._s3

    def _exists(self) -> Tuple[bool, bool]:
        """
        Need to have a way to get bucket and object existence
        seperately for writing.
        """
        s3 = self._s3

        try:
            # s3.meta.client.head_bucket(Bucket=self.bucket)
            logger.info("Checking existence of bucket: %s", self._bucket)
            s3.head_bucket(Bucket=self._bucket)
            logger.info("Bucket: %s exists", self._bucket)
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = int(e.response["Error"]["Code"])
            if error_code == 404:
                logger.info("Bucket: %s does not exist", self._bucket)
                return False, False
            elif error_code == 403:
                logger.info("Access to bucket %s is forbidden", self._bucket)
                return False, False
            elif error_code > 300:
                logger.info(
                    "Error code %s when accessing bucket %s",
                    error_code,
                    self._bucket,
                )
        try:
            logger.info("Checking existence of object: %s", self._key)
            s3.head_object(Bucket=self._bucket, Key=self._key)
            logger.info("Object: %s exists", self._key)
            return True, True
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the object does not exist.
            error_code = int(e.response["Error"]["Code"])
            if error_code == 404:
                logger.info("Object: %s does not exist", self._key)
                return True, False
            else:
                raise RuntimeError("Error occured in Client: %s" % e.response) from e

    @overrides
    def exists(self) -> bool:
        bucket_exists, object_exists = self._exists()
        if bucket_exists and object_exists:
            return True
        else:
            return False

    @overrides
    def delete(self):
        if self.exists():
            self._s3.delete_object(Bucket=self._bucket, Key=self._key)
            return 0
        else:
            return ErrorIO()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module containing bash-related AppDrops

The module contains four classes that offer running bash commands in different
execution modes; that is, in fully batch mode, or with its input and/or output
as a stream of data to the previous/next application.
"""

import contextlib
import os
import signal
import socket
import struct
import subprocess
import tempfile
import threading
import time
import types
import json

from dlg.drop import track_current_drop
from dlg import droputils, utils
from dlg.named_port_utils import (
    DropParser,
    replace_named_ports,
)
from dlg.ddap_protocol import AppDROPStates, DROPStates
from dlg.apps.app_base import BarrierAppDROP, AppDROP
from dlg.exceptions import InvalidDropException, BashAppRuntimeError
from dlg.meta import (
    dlg_string_param,
    dlg_component,
    dlg_batch_input,
    dlg_batch_output,
    dlg_streaming_input,
    dlg_enum_param,
)

# from dlg.runtime.dlg_logging import logproxy, getDLGLogger
import logging
logger = logging.getLogger(f"dlg.{__name__}")



def message_stdouts(prefix, stdout, stderr, enc="utf8"):
    msg = prefix
    if not stdout and not stderr:
        return msg
    msg += ", output follows:"
    if stdout:
        msg += "\n==STDOUT==\n" + utils.b2s(stdout, enc)
    if stderr:
        msg += "\n==STDERR==\n" + utils.b2s(stderr, enc)
    return msg


def close_and_remove(fo, fname):
    fo.close()
    try:
        os.remove(fname)
        logger.debug("Removed %s", fname)
    except OSError:
        logger.exception("Error while removing %s", fname)


def prepare_output_channel(this_node, out_drop):
    """
    Prepares an output channel that will serve as the stdout of a bash command.
    Depending on the values of ``this_node`` and ``out_drop`` the channel will
    be a named pipe or a socket.
    """

    # If the output drop is local then we set up a named pipe
    # otherwise we set up a socket server on our side,
    # which will result in a socket client on the other side
    pipe_name = None
    if out_drop.node == this_node:
        pipe_name = tempfile.mktemp()
        os.mkfifo(pipe_name)
        logger.debug("Created named pipe %s", pipe_name)

        # the pipe needs to be opened after the data is sent to the other
        # application because open() blocks until the other end is also
        # opened
        data = ("pipe://%s" % (pipe_name,)).encode("utf8")
        out_drop.write(data)
        return open(pipe_name, "wb")

    else:
        host = this_node or socket.gethostname()
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind((host, 0))
        sock.listen(1)
        port = sock.getsockname()[1]
        logger.debug("Created TCP socket server at %s:%d", host, port)

        # to get a connection from the other side we have to write the data
        # into the output drop first so the other side connects to us
        out_drop.write(("tcp://%s:%d" % (host, port)).encode("utf8"))
        with contextlib.closing(sock):
            csock, csockaddr = sock.accept()
            csock.setsockopt(
                socket.SOL_SOCKET, socket.SO_LINGER, struct.pack("ii", 1, 1000)
            )
            logger.debug("Received connection from %r", csockaddr)
            return csock


def prepare_input_channel(data):
    """
    Prepares an input channel that will serve as the stdin of a bash command.
    Depending on the contents of ``data`` the channel will be a named pipe or a
    socket.
    """

    # We don't even look at "data", we simply set up a communication channel
    if data.startswith(b"pipe://"):
        pipe_name = data[7:]
        pipe = open(pipe_name, "rb")
        logger.debug("Opened pipe %s for reading", pipe_name)

        # Return an object that runs pipe.read on read()
        # and close_and_remove on close()
        chan = lambda: None  # a simple object where attributes can be set!
        chan.read = pipe.read
        chan.fileno = pipe.fileno
        chan.close = types.MethodType(lambda s: close_and_remove(pipe, pipe_name), chan)
        return chan

    elif data.startswith(b"tcp://"):
        host, port = data[6:].split(b":")
        host = host.decode("utf-8")
        port = int(port)
        sock = utils.connect_to(host, port, 10)
        sock.settimeout(None)
        logger.debug("Connected to TCP socket %s:%d for reading", host, port)
        return sock

    raise RuntimeError(f"Unsupported streaming channel: {data}")


class BashShellBase:
    """
    Common class for BashShell apps. It simply requires a command to be
    specified.
    """

    # TODO: use the shlex module for most of the construction of the
    # command line to get a proper and safe shell syntax
    command = dlg_string_param("Bash command", None)
    input_parser: DropParser = dlg_enum_param(DropParser, "input_parser", DropParser.PICKLE)  # type: ignore

    def initialize(self, **kwargs):
        super(BashShellBase, self).initialize(**kwargs)

        self.proc = None
        self._applicationArgs = self._popArg(kwargs, "applicationArgs", {})

        if not self.command:
            self.command = self._popArg(kwargs, "command", None)
            if not self.command:
                raise InvalidDropException(
                    self, "No command specified, cannot create BashShellApp"
                )

        self.appArgs = self._applicationArgs
        self._recompute_data = {}

    # @track_current_drop
    def _run_bash(self, inputs, outputs, stdin=None, stdout=subprocess.PIPE):
        """
        Runs the given `cmd`. If any `inputs` and/or `outputs` are given
        (dictionaries of uid:drop elements) they are used to replace any placeholder
        value in `cmd` with either drop paths or dataURLs.

        `stdin` indicates at file descriptor or file object to use as the standard
        input of the bash process. If not given no stdin is given to the process.

        Similarly, `stdout` is a file descriptor or file object where the standard
        output of the process is piped to. If not given it is consumed by this
        method and potentially logged.
        """
        logger.debug("Parameters found: %s", json.dumps(self.parameters))
        logger.debug("Bash Inputs: %s; Bash Outputs: %s", inputs, outputs)
        # deal with named ports
        inport_names = self.parameters["inputs"] if "inputs" in self.parameters else []
        outport_names = (
            self.parameters["outputs"] if "outputs" in self.parameters else []
        )

        cmd = self.command.strip()
        cmd = droputils.replace_placeholders(cmd, inputs, outputs)

        keyargs, pargs = replace_named_ports(
            inputs.items(),
            outputs.items(),
            inport_names,
            outport_names,
            self.appArgs,
        )

        for key, value in keyargs.items():
            cmd = cmd.replace(f"{{{key}}}", str(value))
        for key, value in pargs.items():
            cmd = cmd.replace(f"{{{key}}}", str(value))

        # Replace inputs/outputs in command line with paths or data URLs

        # Pass down daliuge-specific information to the subprocesses as environment variables
        env = os.environ.copy()
        env.update({"DLG_UID": self._uid})
        if self.dlg_session_id:
            env.update({"DLG_SESSION_ID": self.dlg_session_id})

        env.update({"DLG_ROOT": utils.getDlgDir()})

        # Wrap everything inside bash
        cmd = ("/bin/bash", "-c", cmd)
        logger.info("Command after wrapping is: %s", cmd)

        start = time.time()

        # Run and wait until it finishes
        process = subprocess.Popen( # pylint: disable=subprocess-popen-preexec-fn
            cmd,
            close_fds=True,
            stdin=stdin,
            stdout=stdout,
            stderr=subprocess.PIPE,
            env=env,
            preexec_fn=os.setsid,
        )
        self.proc = process

        logger.debug("Process launched, waiting now...")

        pstdout, pstderr = process.communicate()
        if stdout != subprocess.PIPE:
            pstdout = b"<piped-out>"
        pcode = process.returncode
        end = time.time()
        logger.info("Finished in %.3f [s] with exit code %d", (end - start), pcode)

        logger.info("Finished in %.3f [s] with exit code %d", (end - start), pcode)
        self._recompute_data["stdout"] = str(pstdout)
        self._recompute_data["stderr"] = str(pstderr)
        self._recompute_data["status"] = str(pcode)
        if pcode == 0 and logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                message_stdouts("Command finished successfully", pstdout, pstderr)
            )
        elif pcode != 0:
            message = "Command didn't finish successfully (exit code %d)" % (pcode,)
            logger.error(message_stdouts(message, pstdout, pstderr))
            raise BashAppRuntimeError(self,message)

    def dataURL(self) -> str:
        return type(self).__name__

    def cancel(self):
        BarrierAppDROP.cancel(self)
        try:
            os.killpg(os.getpgid(self.proc.pid), signal.SIGTERM)
        except OSError:
            logger.exception("Error while terminating process %r", self.proc)

    def generate_recompute_data(self):
        self._recompute_data["command"] = self.command
        return self._recompute_data


class StreamingInputBashAppBase(BashShellBase, AppDROP):
    """
    Base class for bash command applications that consume a stream of incoming
    data.
    """

    def initialize(self, **kwargs):
        BashShellBase.initialize(self, **kwargs)
        self._this_finished = False
        self._prev_finished = False
        self._notified = False

    def notify_if_finished(self):
        if not self._notified and self._prev_finished and self._this_finished:
            self._notified = True
            self._notifyAppIsFinished()

    def dropCompleted(self, uid, drop_state):
        self._prev_finished = True
        self.notify_if_finished()

    def dataWritten(self, uid, data):
        threading.Thread(target=self.execute, args=(data,)).start()

    def execute(self, data):
        logger.debug("Received incoming data connection info: %s", data)
        self.execStatus = AppDROPStates.RUNNING
        try:
            self.run(data)
            drop_state = DROPStates.COMPLETED
            execStatus = AppDROPStates.FINISHED
        except Exception: #pylint: disable=broad-exception-caught
            logger.exception("Error while executing %r", self)
            drop_state = DROPStates.ERROR
            execStatus = AppDROPStates.ERROR
        finally:
            self.execStatus = execStatus
            self.status = drop_state
            self._this_finished = True
            self.notify_if_finished()


#
# Now the actual 4 classes:
# * batch
# * output-only stream
# * input-only stream
# * full-stream
#
##
# @brief BashShellApp
# @details An application component able to run an arbitrary command within the Bash Shell
# @par EAGLE_START
# @param category BashShellApp
# @param tag template
# @param command /String/ComponentParameter/NoPort/ReadWrite//False/False/The command to be executed
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.bash_shell_app.BashShellApp/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param base_name bash_shell_app/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param input_error_threshold 0/Integer/ComponentParameter/NoPort/ReadWrite//False/False/the allowed failure rate of the inputs (in percent), before this component goes to ERROR state and is not executed
# @param n_tries 1/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Specifies the number of times the 'run' method will be executed before finally giving up
# @par EAGLE_END
class BashShellApp(BashShellBase, BarrierAppDROP):
    """
    An app that runs a bash command in batch mode; that is, it waits until all
    its inputs are COMPLETED. It also *doesn't* output a stream of data; see
    StreamingOutputBashApp for those cases.
    """

    component_meta = dlg_component(
        "BashShellApp",
        "An app that runs a bash command in batch mode",
        [dlg_batch_input("text/*", [])],
        [dlg_batch_output("text/*", [])],
        [dlg_streaming_input("text/*")],
    )

    @track_current_drop
    def run(self):
        self._run_bash(self._inputs, self._outputs)


class StreamingOutputBashApp(BashShellBase, BarrierAppDROP):
    """
    Like BashShellApp, but its stdout is a stream of data that is fed into the
    next application.
    """

    component_meta = dlg_component(
        "StreamingOutputBashApp",
        "Like BashShellApp, but its stdout is a stream "
        "of data that is fed into the next application.",
        [dlg_batch_input("text/*", [])],
        [dlg_batch_output("text/*", [])],
        [dlg_streaming_input("text/*")],
    )

    def run(self):
        with contextlib.closing(
            prepare_output_channel(self.node, self.outputs[0])
        ) as outchan:
            self._run_bash(self._inputs, {}, stdout=outchan)
        logger.debug("Closed output channel")


class StreamingInputBashApp(StreamingInputBashAppBase):
    """
    An app that runs a bash command that consumes data from stdin.

    The streaming of data that appears on stdin takes place outside the
    framework; what is streamed through the framework is the information needed
    to establish the streaming channel. This information is also used to kick
    this application off.
    """

    component_meta = dlg_component(
        "StreamingInputBashApp",
        "An app that runs a bash command that consumes data from stdin.",
        [dlg_batch_input("text/*", [])],
        [dlg_batch_output("text/*", [])],
        [dlg_streaming_input("text/*")],
    )

    def run(self, data):
        with contextlib.closing(prepare_input_channel(data)) as inchan:
            self._run_bash({}, self._outputs, stdin=inchan)
        logger.debug("Closed input channel")


class StreamingInputOutputBashApp(StreamingInputBashAppBase):
    """
    Like StreamingInputBashApp, but its stdout is also a stream of data that is
    fed into the next application.
    """

    component_meta = dlg_component(
        "StreamingInputOutputBashApp",
        "Like StreamingInputBashApp, but its stdout is also a "
        "stream of data that is fed into the next application.",
        [dlg_batch_input("text/*", [])],
        [dlg_batch_output("text/*", [])],
        [dlg_streaming_input("text/*")],
    )

    def run(self, data):
        with contextlib.closing(prepare_input_channel(data)) as inchan:
            with contextlib.closing(
                prepare_output_channel(self.node, self.outputs[0])
            ) as outchan:
                self._run_bash({}, {}, stdout=outchan, stdin=inchan)
            logger.debug("Closed output channel")
        logger.debug("Closed input channel")



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import os, string
from abc import abstractmethod

# ===================
# Deployment defaults
# ====================
USER = ""
ACCOUNT = ""
LOGIN_NODE = ""
HOME_DIR = os.environ["HOME"] if "HOME" in os.environ else ""
DLG_ROOT = f"{HOME_DIR}/dlg"
LOG_DIR = f"{DLG_ROOT}/log"
MODULES = ""
VENV = f"{DLG_ROOT}/venv"
DEFAULT_MON_HOST = "dlg-mon.icrar.org"  # TODO: need to get this running
DEFAULT_MON_PORT = 8898
EXEC_PREFIX = "srun -l"


__sub_tpl_str = """#!/bin/bash --login

#SBATCH --nodes=$NUM_NODES
#SBATCH --ntasks-per-node=1
#SBATCH --cpus-per-task=2
#SBATCH --job-name=DALiuGE-$SESSION_ID
#SBATCH --time=$JOB_DURATION
#SBATCH --error=err-%j.log
$MODULES
export DLG_ROOT=$DLG_ROOT
$VENV

$EXEC_PREFIX $PY_BIN -m dlg.deploy.start_dlg_cluster --log_dir $LOG_DIR $GRAPH_PAR $PROXY_PAR $GRAPH_VIS_PAR $LOGV_PAR $ZERORUN_PAR $MAXTHREADS_PAR $SNC_PAR $NUM_ISLANDS_PAR $ALL_NICS $CHECK_WITH_SESSION --ssid $SESSION_ID
"""

__slurm_shebang = "#!/bin/bash --login"

dlg_exec_str = (
    "$EXEC_PREFIX $PY_BIN -m dlg.deploy.start_dlg_cluster" 
    " --log_dir $LOG_DIR $GRAPH_PAR $PROXY_PAR $GRAPH_VIS_PAR $LOGV_PAR $ZERORUN_PAR"
    " $MAXTHREADS_PAR $SNC_PAR $NUM_ISLANDS_PAR $ALL_NICS $CHECK_WITH_SESSION"
    " --ssid $SESSION_ID"
)

init_tpl = string.Template(__sub_tpl_str)


class DefaultConfig(object):

    def __init__(self, user=None):
        self._dict = dict()
        if user:
            print(f"Setting user to {user}")
            self._dict["user"] = user
        self.setpar("host", "LOGIN_NODE")
        self.setpar("account", "ACCOUNT")
        self.setpar("home_dir", "HOME_DIR")
        self.setpar("dlg_root", "DLG_ROOT")
        self.setpar("log_root", "LOG_DIR")
        self.setpar("modules", "MODULES")
        self.setpar("venv", "VENV")
        self.setpar("exec_prefix", "EXEC_PREFIX")

    @abstractmethod
    def init_list(self):
        pass

    def setpar(self, k, v):
        if hasattr(self, v):
            value = getattr(self, v)
            if "user" in self._dict:
                pardict = {"USER": self._dict["user"]}
                value = string.Template(value).safe_substitute(pardict)
            self._dict[k] = value
        else:
            print(f"default[{v}] = '{globals()[v]}'")
            self._dict[k] = globals()[v]

    def getpar(self, k):
        return self._dict.get(k)


#############################
class ICRARHyadesConfig(DefaultConfig):
    MODULES = """
    """
    # The following is more a workaround than a solution
    # requires the user to have a venv exectly in that place
    LOGIN_NODE = "hyades.icrar.org"
    HOME_DIR = "/home/$USER"
    DLG_ROOT = "/home/$USER/dlg"
    LOG_DIR = "/home/$USER/dlg/log"
    VENV = "source /home/$USER/dlg/venv/bin/activate"
    EXEC_PREFIX = ""

    def __init__(self, user=None):
        super(ICRARHyadesConfig, self).__init__(user=user)

    def init_list(self):  # TODO please fill in
        return [self.ACCOUNT, self.LOG_DIR, self.MODULES, self.VENV]


class ICRARoodConfig(DefaultConfig):
    MODULES = """
    module load python/3.8.12
    """
    # The following is more a workaround than a solution
    # requires the user to have a venv exectly in that place
    LOGIN_NODE = "hyades.icrar.org"
    HOME_DIR = os.environ["HOME"] if "HOME" in os.environ else ""
    DLG_ROOT = f"{HOME_DIR}/dlg"
    LOG_DIR = f"{DLG_ROOT}/log"
    VENV = f"source {HOME_DIR}/dlg/venv/bin/activate"

    def __init__(self, user=None):
        super(ICRARoodConfig, self).__init__(user=user)

    def init_list(self):  # TODO please fill in
        return [self.ACCOUNT, self.LOG_DIR, self.MODULES, self.VENV]


class ICRARoodCldConfig(DefaultConfig):
    # The following is more a workaround than a solution
    # requires the user to have a venv exectly in that place
    ACCOUNT = os.environ["USER"]
    HOME_DIR = os.environ["HOME"]
    DLG_ROOT = f"{HOME_DIR}/dlg"
    LOG_DIR = f"{DLG_ROOT}/log"
    # The compute nodes have have required python and DALiuGE but just in case....
    VENV = f"source {DLG_ROOT}/venv/bin/activate"

    def __init__(self, user=None):
        super(ICRARoodCldConfig, self).__init__(user=user)

    def init_list(self):  # TODO please fill in
        return [self.ACCOUNT, self.LOG_DIR, self.VENV]


class GalaxyMWAConfig(DefaultConfig):
    def __init__(self, user=None):
        super(GalaxyMWAConfig, self).__init__(user=user)

    def init_list(self):
        return ["mwaops", "/group/mwaops/cwu/dfms/logs"]


class GalaxyASKAPConfig(DefaultConfig):
    MODULES = """
module swap PrgEnv-cray PrgEnv-gnu
module load python/2.7.10
module load mpi4py
"""
    VENV = ""

    def __init__(self, user=None):
        super(GalaxyASKAPConfig, self).__init__(user=user)

    def init_list(self):
        return [
            "astronomy856",
            "/group/astronomy856/cwu/dfms/logs",
            self.MODULES,
        ]


class MagnusConfig(DefaultConfig):
    def __init__(self, user=None):
        super(MagnusConfig, self).__init__(user=user)

    def init_list(self):
        return ["pawsey0129", "/group/pawsey0129/daliuge_logs"]


class Setonix411Config(DefaultConfig):
    """
    Configuration for project 0411 on Setonix.
    """

    LOGIN_NODE = "setonix.pawsey.org.au"
    ACCOUNT = "pawsey0411"
    USER = os.environ["USER"] if "USER" in os.environ else ""
    HOME_DIR = f"/scratch/{ACCOUNT}"
    DLG_ROOT = f"{HOME_DIR}/$USER/dlg"
    LOG_DIR = f"{DLG_ROOT}/log"
    MODULES = ""
    VENV = f"source /software/projects/{ACCOUNT}/venv/bin/activate"

    MODULES = """module use /group/askap/modulefiles
module load singularity/4.1.0-mpi
module load py-mpi4py/3.1.5-py3.11.6
module load py-numpy/1.26.4"""

    def __init__(self, user=None):
        super(Setonix411Config, self).__init__(user=user)

    def init_list(self):
        return [self.ACCOUNT, f"{self.HOME_DIR}/logs"]


class TianHe2Config(DefaultConfig):
    def __init__(self):
        super(TianHe2Config, self).__init__()

    def init_list(self):  # TODO please fill in
        return ["SHAO", "/group/shao/daliuge_logs"]


##########################################


class ConfigFactory:
    mapping = {
        "galaxy_mwa": GalaxyMWAConfig,
        "galaxy_askap": GalaxyASKAPConfig,
        "magnus": MagnusConfig,
        "galaxy": GalaxyASKAPConfig,
        "setonix": Setonix411Config,
        "shao": TianHe2Config,
        "hyades": ICRARHyadesConfig,
        "ood": ICRARoodConfig,
        "ood_cloud": ICRARoodCldConfig,
    }

    @staticmethod
    def available():
        return list(ConfigFactory.mapping.keys())

    @staticmethod
    def create_config(facility=None, user=None):
        facility = facility.lower() if (facility is not None) else facility
        config = ConfigFactory.mapping.get(facility)(user=user)
        return config
